@@ -3,10 +3,9 @@ import warnings
 from pathlib import Path
 from copy import deepcopy
 from itertools import chain
-from .header import Header
-from .schema import Schema
+from .resource import Resource
 from .system import system
-from .file import File
+from .header import Header
 from .row import Row
 from . import exceptions
 from . import errors
@@ -45,7 +44,7 @@ class Table:
         format? (str): File source's format (csv, xls, ...).
             If not set, it'll be inferred from `source`.
 
-        encoding? (str): An algorithm to hash data.
+        hashing? (str): An algorithm to hash data.
             It defaults to 'md5'.
 
         encoding? (str): Source encoding.
@@ -106,7 +105,7 @@ class Table:
             For more information, please check "Describing  Data" guide.
             It defaults to `['']`
 
-        on_error? (ignore|warn|raise): Define behaviour if there is an error in the
+        onerror? (ignore|warn|raise): Define behaviour if there is an error in the
             header or rows during the reading rows process.
             It defaults to `ignore`.
 
@@ -137,13 +136,14 @@ class Table:
         schema=None,
         sync_schema=False,
         patch_schema=False,
+        # Infer
         infer_type=None,
         infer_names=None,
         infer_volume=config.DEFAULT_INFER_VOLUME,
         infer_confidence=config.DEFAULT_INFER_CONFIDENCE,
         infer_missing_values=config.DEFAULT_MISSING_VALUES,
         # Integrity
-        on_error="ignore",
+        onerror="ignore",
         lookup=None,
     ):
 
@@ -167,7 +167,6 @@ class Table:
         # Store state
         self.__parser = None
         self.__sample = None
-        self.__schema = None
         self.__header = None
         self.__data_stream = None
         self.__row_stream = None
@@ -177,7 +176,6 @@ class Table:
         self.__sample_positions = None
 
         # Store params
-        self.__init_schema = schema
         self.__sync_schema = sync_schema
         self.__patch_schema = patch_schema
         self.__infer_type = infer_type
@@ -185,15 +183,12 @@ class Table:
         self.__infer_volume = infer_volume
         self.__infer_confidence = infer_confidence
         self.__infer_missing_values = infer_missing_values
-        self.__on_error = on_error
+        self.__onerror = onerror
         self.__lookup = lookup
 
-        # Set error handler
-        self.on_error = on_error
-
-        # Create file
-        self.__file = File(
-            source=source,
+        # Create resource
+        self.__resource = Resource.from_source(
+            source,
             scheme=scheme,
             format=format,
             hashing=hashing,
@@ -203,11 +198,14 @@ class Table:
             control=control,
             dialect=dialect,
             query=query,
+            schema=schema,
+            onerror=onerror,
+            trusted=True,
         )
 
     def __setattr__(self, name, value):
-        if name == "on_error":
-            self.__on_error = value
+        if name == "onerror":
+            self.__onerror = value
             return
         super().__setattr__(name, value)
 
@@ -229,7 +227,15 @@ class Table:
         Returns:
             str: file path
         """
-        return self.__file.path
+        return self.__resource.path
+
+    @property
+    def data(self):
+        """
+        Returns:
+            str: file data
+        """
+        return self.__resource.data
 
     @property
     def source(self):
@@ -237,7 +243,7 @@ class Table:
         Returns:
             any: file source
         """
-        return self.__file.source
+        return self.__resource.source
 
     @property
     def scheme(self):
@@ -245,7 +251,7 @@ class Table:
         Returns:
             str?: file scheme
         """
-        return self.__file.scheme
+        return self.__resource.scheme
 
     @property
     def format(self):
@@ -253,7 +259,7 @@ class Table:
         Returns:
             str?: file format
         """
-        return self.__file.format
+        return self.__resource.format
 
     @property
     def hashing(self):
@@ -261,7 +267,7 @@ class Table:
         Returns:
             str?: file hashing
         """
-        return self.__file.hashing
+        return self.__resource.hashing
 
     @property
     def encoding(self):
@@ -269,7 +275,7 @@ class Table:
         Returns:
             str?: file encoding
         """
-        return self.__file.encoding
+        return self.__resource.encoding
 
     @property
     def compression(self):
@@ -277,7 +283,7 @@ class Table:
         Returns:
             str?: file compression
         """
-        return self.__file.compression
+        return self.__resource.compression
 
     @property
     def compression_path(self):
@@ -285,7 +291,7 @@ class Table:
         Returns:
             str?: file compression path
         """
-        return self.__file.compression_path
+        return self.__resource.compression_path
 
     @property
     def control(self):
@@ -293,7 +299,7 @@ class Table:
         Returns:
             Control?: file control
         """
-        return self.__file.control
+        return self.__resource.control
 
     @property
     def query(self):
@@ -301,7 +307,7 @@ class Table:
         Returns:
             Query?: table query
         """
-        return self.__file.query
+        return self.__resource.query
 
     @property
     def dialect(self):
@@ -309,7 +315,7 @@ class Table:
         Returns:
             Dialect?: table dialect
         """
-        return self.__file.dialect
+        return self.__resource.dialect
 
     @property
     def schema(self):
@@ -317,25 +323,32 @@ class Table:
         Returns:
             Schema?: table schema
         """
-        return self.__schema
+        return self.__resource.schema
 
     @property
-    def on_error(self):
+    def stats(self):
+        """Table stats
+
+        The stats object has:
+            - hash: str - hashing sum
+            - bytes: int - number of bytes
+            - fields: int - number of fields
+            - rows: int - number of rows
+
+        Returns:
+            dict?: table stats
+
+        """
+        return self.__resource.stats
+
+    @property
+    def onerror(self):
         """
         Returns:
             ignore|warn|raise: on error bahaviour
         """
-        assert self.__on_error in ["ignore", "warn", "raise"]
-        return self.__on_error
-
-    # TODO: use property wrapper to make it shorter
-    @on_error.setter
-    def on_error(self, value):
-        """
-        Parameters:
-            value (ignore|warn|raise): on error bahaviour
-        """
-        self.__on_error = value
+        assert self.__onerror in ["ignore", "warn", "raise"]
+        return self.__onerror
 
     @property
     def header(self):
@@ -358,21 +371,6 @@ class Table:
         """
         return self.__sample
 
-    @property
-    def stats(self):
-        """Table stats
-
-        The stats object has:
-            - hash: str - hashing sum
-            - bytes: int - number of bytes
-            - rows: int - number of rows
-
-        Returns:
-            dict?: table stats
-
-        """
-        return self.__file.stats
-
     @property
     def data_stream(self):
         """Data stream in form of a generator of data arrays
@@ -400,12 +398,12 @@ class Table:
             FrictionlessException: any exception that occurs
         """
         self.close()
-        if self.__file.query.metadata_errors:
-            error = self.__file.query.metadata_errors[0]
+        if self.__resource.query.metadata_errors:
+            error = self.__resource.query.metadata_errors[0]
             raise exceptions.FrictionlessException(error)
         try:
-            self.__file.stats = {"hash": "", "bytes": 0, "rows": 0}
-            self.__parser = system.create_parser(self.__file)
+            self.__resource.stats = {"hash": "", "bytes": 0, "fields": 0, "rows": 0}
+            self.__parser = system.create_parser(self.__resource)
             self.__parser.open()
             self.__data_stream = self.__read_data_stream()
             self.__row_stream = self.__read_row_stream()
@@ -416,7 +414,7 @@ class Table:
             self.close()
             # Ensure not found file is a scheme error
             if exception.error.code == "format-error":
-                loader = system.create_loader(self.__file)
+                loader = system.create_loader(self.__resource)
                 loader.open()
                 loader.close()
             raise
@@ -439,9 +437,6 @@ class Table:
         """
         return self.__parser is None
 
-        # Read
-        return self.__on_error
-
     def read_data(self):
         """Read data stream into memory
 
@@ -456,9 +451,8 @@ class Table:
         return self.__read_data_stream_create()
 
     def __read_data_stream_create(self):
-        stats = self.__file.stats
-        limit = self.__file.query.limit_rows
-        offset = self.__file.query.offset_rows or 0
+        limit = self.__resource.query.limit_rows
+        offset = self.__resource.query.offset_rows or 0
         sample_iterator = self.__read_data_stream_create_sample_iterator()
         parser_iterator = self.__read_data_stream_create_parser_iterator()
         for row_position, cells in chain(sample_iterator, parser_iterator):
@@ -467,9 +461,9 @@ class Table:
                 offset -= 1
                 continue
             self.__row_number += 1
-            stats["rows"] = self.__row_number
+            self.__resource.stats["rows"] = self.__row_number
             yield cells
-            if limit and limit <= stats["rows"]:
+            if limit and limit <= self.__resource.stats["rows"]:
                 break
 
     def __read_data_stream_create_sample_iterator(self):
@@ -490,7 +484,6 @@ class Table:
         header = []
         field_positions = []
         sample_positions = []
-        schema = Schema(self.__init_schema)
 
         # Prepare header
         buffer = []
@@ -504,7 +497,7 @@ class Table:
 
         # Infer header
         row_number = 0
-        dialect = self.__file.dialect
+        dialect = self.__resource.dialect
         if dialect.get("header") is None and dialect.get("headerRows") is None and widths:
             dialect["header"] = False
             width = round(sum(widths) / len(widths))
@@ -550,6 +543,7 @@ class Table:
                     break
 
         # Infer schema
+        schema = self.__resource.schema
         if not schema.fields:
             schema.infer(
                 sample,
@@ -575,20 +569,20 @@ class Table:
             for field in schema.fields:
                 field.update((fields.get(field.get("name"), {})))
 
-        # Confirm schema
+        # Validate schema
         if len(schema.field_names) != len(set(schema.field_names)):
             note = "Schemas with duplicate field names are not supported"
             raise exceptions.FrictionlessException(errors.SchemaError(note=note))
 
-        # Store state
+        # Store state/stats
         self.__sample = sample
-        self.__schema = schema
         self.__field_positions = field_positions
         self.__sample_positions = sample_positions
         self.__header = Header(header, schema=schema, field_positions=field_positions)
+        self.__resource.stats["fields"] = len(schema.fields)
 
     def __read_data_stream_infer_header(self, header_data):
-        dialect = self.__file.dialect
+        dialect = self.__resource.dialect
 
         # No header
         if not dialect.header:
@@ -610,8 +604,8 @@ class Table:
         # Filter header
         filter_header = []
         field_positions = []
-        limit = self.__file.query.limit_fields
-        offset = self.__file.query.offset_fields or 0
+        limit = self.__resource.query.limit_fields
+        offset = self.__resource.query.offset_fields or 0
         for field_position, header in enumerate(header, start=1):
             if self.__read_data_stream_pick_skip_field(field_position, header):
                 if offset:
@@ -628,9 +622,9 @@ class Table:
         match = True
         for name in ["pick", "skip"]:
             if name == "pick":
-                items = self.__file.query.pick_fields_compiled
+                items = self.__resource.query.pick_fields_compiled
             else:
-                items = self.__file.query.skip_fields_compiled
+                items = self.__resource.query.skip_fields_compiled
             if not items:
                 continue
             match = match and name == "skip"
@@ -651,9 +645,9 @@ class Table:
         cell = "" if cell is None else str(cell)
         for name in ["pick", "skip"]:
             if name == "pick":
-                items = self.__file.query.pick_rows_compiled
+                items = self.__resource.query.pick_rows_compiled
             else:
-                items = self.__file.query.skip_rows_compiled
+                items = self.__resource.query.skip_rows_compiled
             if not items:
                 continue
             match = match and name == "skip"
@@ -671,7 +665,7 @@ class Table:
         return match
 
     def __read_data_stream_filter_data(self, cells, field_positions):
-        if self.__file.query.is_field_filtering:
+        if self.__resource.query.is_field_filtering:
             result = []
             for field_position, cell in enumerate(cells, start=1):
                 if field_position in field_positions:
@@ -702,9 +696,9 @@ class Table:
         # Handle header errors
         if not self.header.valid:
             error = self.header.errors[0]
-            if self.__on_error == "warn":
+            if self.onerror == "warn":
                 warnings.warn(error.message, UserWarning)
-            elif self.__on_error == "raise":
+            elif self.onerror == "raise":
                 raise exceptions.FrictionlessException(error)
 
         # Create state
@@ -728,10 +722,10 @@ class Table:
             # Create row
             row = Row(
                 cells,
-                schema=self.__schema,
+                schema=self.__resource.schema,
                 field_positions=self.__field_positions,
                 row_position=self.__row_position,
-                row_number=self.__file.stats["rows"],
+                row_number=self.__resource.stats["rows"],
             )
 
             # Unique Error
@@ -780,9 +774,9 @@ class Table:
             # Handle row errors
             if not row.valid:
                 error = row.errors[0]
-                if self.__on_error == "warn":
+                if self.onerror == "warn":
                     warnings.warn(error.message, UserWarning)
-                elif self.__on_error == "raise":
+                elif self.onerror == "raise":
                     raise exceptions.FrictionlessException(error)
 
             # Stream row
@@ -796,10 +790,10 @@ class Table:
     # Write
 
     # NOTE: implement proper usage of loaders (e.g. write to s3)
-    # NOTE: allow None target and return result for inline/pandas/etc?
+    # TODO: allow None target and return result for inline/pandas/etc?
     def write(
         self,
-        target,
+        target=None,
         *,
         scheme=None,
         format=None,
@@ -818,8 +812,8 @@ class Table:
         """
 
         # Create file
-        file = File(
-            source=target,
+        resource = Resource.from_source(
+            target,
             scheme=scheme,
             format=format,
             hashing=hashing,
@@ -828,12 +822,15 @@ class Table:
             compression_path=compression_path,
             control=control,
             dialect=dialect,
+            schema=self.schema,
+            trusted=True,
         )
 
         # Write file
         row_stream = self.__write_row_stream_create()
-        parser = system.create_parser(file)
+        parser = system.create_parser(resource)
         parser.write(row_stream)
+        return resource.source
 
     def __write_row_stream_create(self):
         self.__read_data_stream_raise_closed()
